
package projekat1;

import com.sun.jdi.VirtualMachine;
import com.sun.jdi.Bootstrap;
import com.sun.jdi.connect.*;

import java.util.Map;
import java.util.List;
import java.util.Iterator;

import java.io.PrintWriter;
import java.io.FileWriter;
import java.io.IOException;


public class Trace {

    // Running remote VM
    private VirtualMachine vm;

    // Thread transferring remote error stream to our error stream
    private Thread errThread = null;

    // Thread transferring remote output stream to our output stream
    private Thread outThread = null;

    // 1 ako je debug mode ukljucen (nije implementirano do kraja, default 0)
    private int debugTraceMode = 0;

    //  true ako se prate promene vrednosti polja
    private boolean watchFields = true;

    // paketi i klase koje se ne zele uzeti u obzir - regularni izrazi
    private String[] excludes = {"java.lang.*", "java.io.*","java.nio.*","java.s*","java.net.*","java.util.H*","java.util.z*",
            "java.*",
            "javax.*", "sun.*", 
				 "com.sun.*"};

    /**
     * main
     */
    public static void main(String[] args) {
        new Trace(args);
    }

    /**
     * Parsira argumente
     * startuje java VM
     * 
     */
    Trace(String[] args) {
        try{
        	PrintWriter writer = new PrintWriter(System.out);
        	int inx;
        	for (inx = 0; inx < args.length; ++inx) {
        	    String arg = args[inx];
        	    if (arg.charAt(0) != '-') {
        		break;
        	    }
        	    if (arg.equals("-output")) {
        		try {
        		    writer = new PrintWriter(new FileWriter(args[++inx]));
        		} catch (IOException exc) {
        		    System.err.println("Cannot open output file: " + args[inx]
        				       + " - " +  exc);
        		    System.exit(1);
        		}
        	    } else if (arg.equals("-all")) {
        		excludes = new String[0];
        	    } else if (arg.equals("-fields")) {
        		watchFields = true;
        	    } else if (arg.equals("-dbgtrace")) {
        		debugTraceMode = Integer.parseInt(args[++inx]);
        	    } else if (arg.equals("-help")) {
        		usage();
        		System.exit(0);
        	    } else {
        		System.err.println("No option: " + arg);
        		usage();
        		System.exit(1);
        	    }
        	}
            if (inx >= args.length) {
        	    System.err.println("<class> missing");
                    usage();
                    System.exit(1);
            }
            StringBuffer sb = new StringBuffer();
            sb.append(args[inx]);
            for (++inx; inx < args.length; ++inx) {
                sb.append(' ');
                sb.append(args[inx]);
            }
            vm = launchTarget(sb.toString());
            generateTrace(writer);
        }catch (Exception e) {
            e.printStackTrace();
        }
    }


    /**
     * Generise trace.
     * 
     */
    void generateTrace(PrintWriter writer) {
        vm.setDebugTraceMode(debugTraceMode);
        EventThread eventThread = new EventThread(vm, excludes, writer); 
        eventThread.setEventRequests(watchFields);
        eventThread.start();
        redirectOutput();
        vm.resume();

        // pocetak shutdown-a VM
	try {
	    eventThread.join();
	    errThread.join(); // Make sure output is forwarded 
	    outThread.join(); // before we exit
	} catch (InterruptedException exc) {
	    // we don't interrupt
	}
	writer.close();
    }   

    /**
     * Startovanje VM
     * 
     */
    VirtualMachine launchTarget(String mainArgs) {
	LaunchingConnector connector = findLaunchingConnector();
	Map arguments = connectorArguments(connector, mainArgs);
        try {
	    return connector.launch(arguments);
        } catch (IOException exc) {
            throw new Error("Unable to launch target VM: " + exc);
        } catch (IllegalConnectorArgumentsException exc) {
            throw new Error("Internal error: " + exc);
        } catch (VMStartException exc) {
            throw new Error("Target VM failed to initialize: " +
			    exc.getMessage());
        }
    }

    void redirectOutput() {
        Process process = vm.process();

        // Kopira output i err odredisnog programa u errThread i outThread
        errThread = new StreamRedirectThread("error reader",
                                             process.getErrorStream(),
                                             System.err);
        outThread = new StreamRedirectThread("output reader",
                                             process.getInputStream(),
                                             System.out);
        errThread.start();
        outThread.start();
    }

    /**
     * Find a com.sun.jdi.CommandLineLaunch connector
     */
    LaunchingConnector findLaunchingConnector() {
        List connectors = Bootstrap.virtualMachineManager().allConnectors();
        Iterator iter = connectors.iterator();
        while (iter.hasNext()) {
            Connector connector = (Connector)iter.next();
            if (connector.name().equals("com.sun.jdi.CommandLineLaunch")) {
                return (LaunchingConnector)connector;
            }
        }
        throw new Error("No launching connector");
    }

    /**
     * Return the launching connector's arguments.
     */
    Map connectorArguments(LaunchingConnector connector, String mainArgs) {
        Map arguments = connector.defaultArguments();
        Connector.Argument mainArg = 
	                   (Connector.Argument)arguments.get("main");
        if (mainArg == null) {
            throw new Error("Bad launching connector");
        }
	mainArg.setValue(mainArgs);

	if (watchFields) { 
	    
	    Connector.Argument optionArg = 
		(Connector.Argument)arguments.get("options");
	    if (optionArg == null) {
		throw new Error("Bad launching connector");
	    }
	    optionArg.setValue("-classic");
	}
	return arguments;
    }

    /**
     * Print command line usage help
     */
    void usage() {
	System.err.println("Upotreba: java Trace <options> <class> <args>");
	System.err.println("<options> su:");
	System.err.println(
"  -output <filename>   output se upisuje u fajl <filename>");
	System.err.println(
"  -all                 u obradu ukljuciti i sistemske klase");
	System.err.println(
"  -help                help");
	System.err.println("<class> odredisni program za trace");
	System.err.println("<args> argumenti <class>");
    }
}
